#include "main.h"

#define THREADS     4

typedef struct pixelainfo { /* pixel array info */
    unsigned int* start;
    unsigned int length;
} pixelainfo_t;

typedef struct countinfo {
    unsigned int start;
    unsigned int length;
    unsigned int ret;
} countinfo_t;

unsigned char* colors;
const int colorLen = 0xffffff + 1;

DWORD WINAPI colorThread(void* data) {
    unsigned int i;
    pixelainfo_t* part = (pixelainfo_t*) data;
    for(i = 0; i < part->length; i++) {
        colors[*(part->start + i) & 0xffffff] = 1;
    }
    return 0;
}

DWORD WINAPI additionThread(void* data) {
    int i;
    countinfo_t* part = (countinfo_t*) data;
    part->ret = 0;
    for(i = 0; i < part->length; i++) {
        part->ret += colors[part->start + i];
    }
    return 0;
}

int DLL_EXPORT getColorCount(unsigned int* pixel, int len) {
    int i, colorSum = 0;
    colors = (unsigned char*) calloc(colorLen, sizeof(unsigned char));
    HANDLE threads[THREADS];
    pixelainfo_t* pixelparts[THREADS];
    countinfo_t* countparts[THREADS];

    for(i = 0; i < THREADS; i++) {
        /* partitioning pixel map and sum map, also start color count threads */
        pixelparts[i] = (pixelainfo_t*) malloc(sizeof(pixelainfo_t));
        pixelparts[i]->start = pixel + i * (len / THREADS);
        pixelparts[i]->length = len / THREADS;
        threads[i] = CreateThread(NULL, 0, colorThread, pixelparts[i], 0, NULL);
        countparts[i] = (countinfo_t*) malloc(sizeof(countinfo_t));
        countparts[i]->start = i * (colorLen / THREADS);
        countparts[i]->length = colorLen / THREADS;
    }
    /* wait for color count threads (JOIN) */
    WaitForMultipleObjects(THREADS, threads, TRUE, INFINITE);
    for(i = 0; i < THREADS; i++) {
        /* close color count threads, start addition threads */
        CloseHandle(threads[i]);
        free(pixelparts[i]);
        threads[i] = CreateThread(NULL, 0, additionThread, countparts[i], 0, NULL);
    }
    /* wait for addition threads (JOIN) */
    WaitForMultipleObjects(THREADS, threads, TRUE, INFINITE);
    for(i = 0; i < THREADS; i++) {
        /* close addition threads, calculate final sum from parts */
        CloseHandle(threads[i]);
        colorSum += countparts[i]->ret;
        free(countparts[i]);
    }
    free(colors);
    return colorSum;
}

DLL_EXPORT BOOL APIENTRY DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
    switch (fdwReason)
    {
        case DLL_PROCESS_ATTACH:
            // attach to process
            // return FALSE to fail DLL load
            break;

        case DLL_PROCESS_DETACH:
            // detach from process
            break;

        case DLL_THREAD_ATTACH:
            // attach to thread
            break;

        case DLL_THREAD_DETACH:
            // detach from thread
            break;
    }
    return TRUE; // succesful
}
